@@ -1,17 +1,14 @@
-import logging
 import os
 import sys
-import warnings
 from functools import partial
 from multiprocessing import Process
 from time import sleep
-from typing import NamedTuple
+from collections import OrderedDict
 
 # We may need below, we may not. What can I say? Dill waters run deep.
-import deprecation
 import dill
 
-from pyqalx.config import BotConfig
+from pyqalx.config import BotConfig, UserConfig
 from pyqalx.core.adapters import QalxQueue, QalxWorker
 from pyqalx.core.session import QalxSession
 from pyqalx.core.errors import (
@@ -19,18 +16,13 @@ from pyqalx.core.errors import (
     QalxStepFunctionNotDefined,
     QalxMultipleEntityReturned,
     QalxEntityNotFound,
+    QalxAPIResponseError,
 )
-from pyqalx.core.errors import QalxError, QalxConfigError
-from pyqalx.core.log import (
-    exception_hook,
-    QalxRotatingFileHandler,
-    QalxLogFormatter,
-)
+from pyqalx.core.errors import QalxError
+from pyqalx.core.log import exception_hook, configure_logging
 from pyqalx.core.signals import QalxSignal
 from pyqalx.factories.factory import Factory
 
-logger = logging.getLogger(__name__)
-
 dill.settings["recurse"] = True
 
 
@@ -73,13 +65,13 @@ class QalxJob:
         "progress",
         "warnings",
         "errors",
-        "worker_logger",
         "_current_step",
         "_process",
         "_worker_entity",
         "_bot_entity",
         "_queue_message",
         "_step_results",
+        "context",
         "do_workflow",
     )
 
@@ -89,18 +81,21 @@ class QalxJob:
         self._bot_entity = worker_process.bot_entity
         self._worker_entity = worker_process.api_entity
         self.session = worker_process.session
-        self.worker_logger = worker_process.logger
 
         self._current_step = "class init"
-        self._step_results = []
+        # We need to preserve the order of step results and although in Python
+        # >= 3.6 this is implemented in the dict class, it cannot be relied upon
+        # https://docs.python.org/3.6/whatsnew/3.6.html#whatsnew36-compactdict
+        self._step_results = OrderedDict()
+        self.context = dict()
         self.do_workflow = True
 
     @property
     def log(self):
         """
-        An instance of `worker_logger` to enable logging from step functions
+        An instance of the workers log to enable logging from step functions
         """
-        return self.worker_logger
+        return self._process.log
 
     @property
     def e(self):
@@ -148,77 +143,34 @@ class QalxJob:
             requeue_job=requeue_job,
         )
 
-    def publish_status(self, status="", status_string="", status_dict=None):
+    def publish_status(self, status):
         """update the status of the worker
 
         :param status: short message about the worker status
         :type status: str
-        :param status_string: short message about the worker status
-        :type status_string: str
-        :param status_dict: structured status data
-        :type status_dict: dict
         """
-        if status_string:  # pragma: no cover
-            the_warning = deprecation.DeprecatedWarning(
-                function="status_string",
-                deprecated_in="0.8",
-                removed_in="1.0",
-                details="`status_string` will be renamed to `status` in "
-                "future versions to maintain consistency with Bots",
-            )
-            warnings.warn(
-                the_warning, category=DeprecationWarning, stacklevel=2
-            )
-            # If the user supplies both arguments through the deprecation
-            # period we won't know which to use
-            assert not all(
-                [status, status_string]
-            ), "Ensure only `status` or `status_string` is specified"
-            if status_string and not status:
-                # Ensure that we use the status they passed through as a kwarg
-                status = status_string
-
-        if status_dict is not None:  # pragma: no cover
-            the_warning = deprecation.DeprecatedWarning(
-                function="status_dict",
-                deprecated_in="0.8",
-                removed_in="1.0",
-                details="Data stored in `status_dict` is not saved by the API."
-                "The argument will be removed in future versions.",
-            )
-            warnings.warn(
-                the_warning, category=DeprecationWarning, stacklevel=2
-            )
         self.worker_adapter.update_status(
             entity=self._worker_entity,
             bot_entity=self._bot_entity,
             status=status,
         )
 
-    def add_step_result(self, success=True, result_data=None):
+    def add_step_result(self, success=True):
         """indicate success and add info about the current step
 
         :param success: was the step successful or not
         :type success: bool
-        :param result_data: data from the result
-        :type result_data: dict
         """
-        self._step_results.append(
-            StepResult(
-                success=success,
-                step_name=self._current_step,
-                result_data=result_data,
-            )
-        )
+        self._step_results[self._current_step] = success
 
     @property
     def last_step_result(self):
         """
-        The last :class:`~pyqalx.bot.StepResult` object (or None if no
-        previous steps) for this job
+        The success flag for the previous step result (or None if no previous
+        steps) for this job
         """
         if self._step_results:
-            return self._step_results[-1]
+            return self._step_results[list(self._step_results)[-1]]
         else:
             return None
 
@@ -230,11 +182,22 @@ class QalxJob:
         :type entity_type: str
         :param entity_guid: the guid of the entity
         :type entity_guid: str
-        :return: a child of pyqalx.core.entities.QalxEntity
+        :return:    a child of pyqalx.core.entities.QalxEntity if the entity is
+                    found on the API, otherwise None
         """
         adapter = getattr(self.session, entity_type)
         self.entity_adapter = adapter
-        self.entity = adapter.get(entity_guid, config=self.session.config)
+        try:
+            self.entity = adapter.get(entity_guid, config=self.session.config)
+        except QalxAPIResponseError:
+            self.log.error(
+                f"The {entity_type} with guid {entity_guid} was not found"
+                + " on qalx. This probably means that it was manually "
+                + f"deleted. The {entity_type} was not processed and the "
+                + "message will be deleted from the queue."
+            )
+            self.delete_message()
+            self.entity = None
         return self.entity
 
     def add_item_data(self, item_key, data, meta=None):
@@ -290,20 +253,7 @@ class QalxJob:
         """
         Helper method for deleting a message from a Queue
         """
-        return self._worker_entity._delete_message()
-
-
-class StepResult(NamedTuple):
-    """
-    An NamedTuple representing the result of a single step function.
-    """
-
-    # The name of the step function that this instance refers to
-    step_name: str = ""
-    # Was this step successful or not?
-    success: bool = False
-    # A dict of data for the result of the step
-    result_data: dict = {}
+        return self._process._delete_message()
 
 
 class Bot:
@@ -377,10 +327,9 @@ class Bot:
         user_config_class=None,
         bot_config_class=None,
         entity_classes=None,
-        user_profile_name="default",
-        bot_profile_name="default",
+        user_profile_name=UserConfig.default,
+        bot_profile_name=BotConfig.default,
         skip_ini=False,
-        add_queue=True,
         stop=False,
         meta=None,
         workflow=None,
@@ -410,9 +359,6 @@ class Bot:
         :type skip_ini: bool
         :param processes: The number of processes this bot should spawn
         :type processes: int
-        :param add_queue: Should the queue be created if it doesn't already
-                          exist
-        :type add_queue: bool
         :param stop:    Should the bot be stopped right after startup. This is
                         essential functionality during factory building
         :type  stop:    bool
@@ -457,26 +403,7 @@ class Bot:
             config_class=bot_config_class,
             skip_ini=skip_ini,
         )
-        if add_queue:
-            # We use `get_or_create` to stop the bot erroring because
-            # the user didn't know the queue already existed
-            the_warning = deprecation.DeprecatedWarning(
-                function="add_queue",
-                deprecated_in="0.7",
-                removed_in="1.0",
-                details="Bots now use `get_or_create` when given a "
-                "`queue_name` so the user no longer has the "
-                "responsibility to create the queue before starting "
-                "the bot",
-            )
-            warnings.warn(
-                the_warning, category=DeprecationWarning, stacklevel=2
-            )
-
-            queue = user_session.queue.get_or_create(name=self.queue_name)
-        else:
-            # We create the queue if it doesn't already exist
-            queue = user_session.queue.get_or_create(name=self.queue_name)
+        queue = user_session.queue.get_or_create(name=self.queue_name)
         # Add the queue_name - this makes it easier to find bots looking at
         # specific queues.
         bot_session.config.update({"queue_name": self.queue_name})
@@ -490,17 +417,19 @@ class Bot:
 
         if self.initialisation_function:
             init_result = self.initialisation_function(user_session)
-            logger.debug(f"Initialisation function returns {init_result}")
+            bot_session.log.debug(
+                f"Initialisation function returns {init_result}"
+            )
             if not init_result:
                 raise QalxBotInitialisationFailed(
                     "The initialisation function returned a False-like value."
                 )
         steps = self._collect_steps()
 
-        # We need to update the config with the Bot Token
-        self.bot_adapter.session._update_config(
-            config={"TOKEN": self.bot_entity["info"]["credentials"]["token"]}
-        )
+        # Update the token on the bot_session to ensure all subsequent
+        # requests use the token from the bot
+        token = self.bot_entity["info"]["credentials"]["token"]
+        bot_session.rest_api.token = token
 
         self.bot_entity = self.bot_adapter.replace_workers(
             bot_entity=self.bot_entity, workers=processes
@@ -524,14 +453,14 @@ class Bot:
                     workflow=self.workflow,
                 )
             )
-            logger.debug(
+            bot_session.log.debug(
                 f"Created worker {worker} to read from {self.queue_name}"
             )
         for p in workers:
             p.start()
         for p in workers:
             p.join()
-        logger.debug("Workers joined.")
+        bot_session.log.debug("Workers joined.")
 
     def _register_entities(self, entity_classes):
         """
@@ -651,8 +580,11 @@ class Worker(Process):
         self.api_entity = api_entity
         self.worker_adapter = worker_adapter
         self.session = self.worker_adapter.session
+        # Reinitialise the session due to mulitprocessing
+        # https://github.com/psf/requests/issues/4323
+        self.session.rest_api._initialise_session()
 
-        self.logger = None
+        self.log = None
         self.queue_adapter = QalxQueue(session=self.session)
         self.stop = False
         self.config = self.session.config
@@ -660,34 +592,18 @@ class Worker(Process):
         self.workflow = workflow
         self.worker_signal = None
 
-    def configure_logging(self, index):
+    def configure_logging(self):
         """
-        Configures the logger for the Worker. Each Worker logs to its own
+        Configures the log for the Worker. Each Worker logs to its own
         file. This prevents issues with multiple processes trying to write
         to the same file
-        :param index: The index of the worker to use in the filename
-        :return: An instance of `logging.Logger`
         """
         directory = self.config["WORKER_LOG_FILE_DIR"]
         level = self.config["LOGGING_LEVEL"]
-        if os.path.isdir(directory) is False:
-            raise QalxConfigError(
-                f"WORKER_LOG_FILE_DIR config value "
-                f"not a valid directory. Got `{directory}`"
-            )
-
-        worker_logger = logging.getLogger(f"pyqalx.worker.{index}")
-
         filename = os.path.join(
-            directory, f'.worker_{self.bot_entity["name"]}.{index}.log'
+            directory, f'.worker_{self.bot_entity["name"]}.{self.index}.log'
         )
-        handler = QalxRotatingFileHandler(filename=filename)
-
-        handler.setFormatter(QalxLogFormatter())
-        handler.setLevel(level)
-        worker_logger.addHandler(handler)
-        worker_logger.setLevel(level)
-        return worker_logger
+        self.log = configure_logging(f"worker.{self.index}", filename, level)
 
     def _reload_signal(self):
         """
@@ -753,7 +669,7 @@ class Worker(Process):
         Workflow step function. Used to handle passing a job to multiple bots
         once all step functions have been completed on the worker
         """
-        self.logger.debug("Attempting worklow")
+        self.log.debug("Attempting worklow")
         factory_namespace_key = Factory.PYQALX_METADATA_NAMESPACE_KEY
         factory_guid_key = Factory.PYQALX_METADATA_FACTORY_GUID_KEY
         factory_plan_bot_name_key = Factory.PYQALX_METADATA_PLAN_BOT_NAME_KEY
@@ -761,7 +677,7 @@ class Worker(Process):
             factory_guid_key, None
         )
         for bot_name in self.workflow:
-            self.logger.debug(
+            self.log.debug(
                 f"Workflow found." f"Next bot in workflow: `{bot_name}`"
             )
             # By default, just try and find the bot based on the name
@@ -784,15 +700,15 @@ class Worker(Process):
                     f"for `{bot_name}` in the defined workflow. "
                     f"Query was `{query}`: `{exc}`"
                 )
-                self.logger.warning(warn_msg)
+                self.log.warning(warn_msg)
                 continue
             queue = self.session.queue.get_by_name(name=bot.config.queue_name)
-            self.logger.debug(
+            self.log.debug(
                 f"Submitting `{self.job.entity}` "
                 f"to `{queue.name}` on workflow bot"
                 f" name `{bot_name}`"
             )
-            queue.submit_entity(self.job.entity)
+            self.job.entity.add_to_queue(self.job.entity, queue)
 
     def _delete_message(self):
         """
@@ -811,15 +727,15 @@ class Worker(Process):
         :return: the next spreading factor to use
         """
 
-        time_to_wait = int(self.config["Q_WAITTIMESECONDS"])
-        max_spreading = int(self.config["Q_WAITTIMEMAXSPREADING"])
+        time_to_wait = self.config["Q_WAITTIMESECONDS"]
+        max_spreading = self.config["Q_WAITTIMEMAXSPREADING"]
         self.job._queue_message = None
         self.job._current_step = "OnWait"
         if self.steps.get("onwait_function") is not None:
             self.steps["onwait_function"](self.job)
         total_wait_seconds = time_to_wait * spread_factor
         sleep(total_wait_seconds)
-        self.logger.debug(f"wait for {total_wait_seconds}s")
+        self.log.debug(f"wait for {total_wait_seconds}s")
         self._check_worker_signal()
         self._check_cold_exit(reload_signal=False)  # we reloaded in prev line
         next_spread = spread_factor + 1
@@ -827,8 +743,8 @@ class Worker(Process):
         return min(max_spreading, next_spread)
 
     def _continue_processing(self):
-        if self.job.last_step_result:
-            return (not self.stop) and self.job.last_step_result.success
+        if self.job.last_step_result is not None:
+            return (not self.stop) and self.job.last_step_result
         else:
             return not self.stop
 
@@ -838,27 +754,26 @@ class Worker(Process):
         refreshes it if necessary
         """
         if self.queue._credentials_expiring_soon():
-            self.logger.debug(
+            self.log.debug(
                 f"refreshing security credentials for queue "
                 f'`{self.queue["name"]}`'
             )
             self.queue = self.queue_adapter.reload(self.queue)
 
     def _run_functions(self):
-        # Each worker has its own dedicated logger
-        self.logger = self.configure_logging(index=self.index)
+        # Each worker has its own dedicated log
         self.job = QalxJob(self)
         _startup = True
         # avoid pickle errors
         self.steps = {}
         for f_name, f_dill in self._steps.items():  # de-serialise functions
-            self.logger.debug(f"try to undill step {f_name}")
+            self.log.debug(f"try to undill step {f_name}")
             try:
                 self.steps[f_name] = dill.loads(f_dill)
             except TypeError:
                 self.steps[f_name] = f_dill
 
-        self.logger.debug("running")
+        self.log.debug("running")
         if self.steps.get("begin_function"):
             self.job._current_step = "Begin"
 
@@ -869,6 +784,9 @@ class Worker(Process):
         kill_after = self.config["KILL_AFTER"]
         spread_factor = 1
         no_message_attempts = 0
+        # Checks the worker signal in case the bot is started in a "stopped"
+        # state
+        self._check_worker_signal()
 
         while True:
             # Check if the credentials need refreshing. They may have expired
@@ -877,19 +795,19 @@ class Worker(Process):
             self._check_queue_credentials()
             if self.stop:
                 spread_factor = self._wait_with_spreading(spread_factor)
-                self.logger.debug(
+                self.log.debug(
                     f"Stop signal on Worker waited with factor {spread_factor}"
                 )
             else:
                 q_msgs = self.queue_adapter.get_messages(worker=self)
-                self.logger.debug(f"got {len(q_msgs)} messages")
+                self.log.debug(f"got {len(q_msgs)} messages")
                 if q_msgs:
                     for q_msg in q_msgs:
                         if not _startup:
                             self.job = QalxJob(self)
                         _startup = False
                         self.job._queue_message = q_msg
-                        self.logger.debug(f"process {q_msg.body}")
+                        self.log.debug(f"process {q_msg.body}")
                         self._process_functions()
                         self._check_worker_signal()
                     no_message_attempts = 0  # reset the attempt counter
@@ -898,9 +816,7 @@ class Worker(Process):
                     spread_factor = self._wait_with_spreading(spread_factor)
                     no_message_attempts += 1
 
-            if (kill_after is not None) and (
-                no_message_attempts >= int(kill_after)
-            ):
+            if (kill_after is not None) and (no_message_attempts >= kill_after):
                 self._do_exit("KILL_AFTER reached", 72593)
 
     def _process_functions(self):
@@ -924,16 +840,16 @@ class Worker(Process):
                                 "entity_guid": self.job.entity["guid"],
                                 "entity_type": self.job.entity.entity_type,
                             },
-                            logger=self.logger,
+                            log=self.log,
                         )
 
                     # check if we need to cold exit after every function
                     self._check_cold_exit()
-                    if self.config.getboolean("SAVE_INTERMITTENT"):
+                    if self.config["SAVE_INTERMITTENT"]:
                         self.job.save_entity()
                 else:
                     break
-            if not self.config.getboolean("SKIP_FINAL_SAVE"):
+            if not self.config["SKIP_FINAL_SAVE"]:
                 self.job.save_entity()
             # `_workflow` is executed only if the flag on the job entity is True
             if self.job.do_workflow is True:
@@ -964,9 +880,8 @@ class Worker(Process):
 
             entity_type = self.job._queue_message.body["entity_type"]
             entity_guid = self.job._queue_message.body["entity_guid"]
-            self.job.get_entity(entity_type, entity_guid)
-
-            process_as_far_as_possible(process_functions)
+            if self.job.get_entity(entity_type, entity_guid):
+                process_as_far_as_possible(process_functions)
         else:
             self._delete_message()
 
@@ -976,6 +891,10 @@ class Worker(Process):
         that may occur with it and pass that to our custom
         uncaught exception handler
         """
+        # This is the earliest the log can be configured to avoid
+        # pickling issues on windows.  Any logging before this needs
+        # to be done using a QalxSession#log instance
+        self.configure_logging()
         try:
             self._run_functions()
         except Exception:
@@ -988,5 +907,5 @@ class Worker(Process):
                 exc_type=QalxError,
                 exc_value=exc_value,
                 exc_traceback=exc_traceback,
-                logger=self.logger,
+                log=self.log,
             )
