@@ -1,11 +1,11 @@
-import json
 import os
 
-import logging
 from types import MethodType
+from bson_extra import bson_extra
 
 from pyqalx.config import UserConfig, BotConfig
 from pyqalx.core.entities import QalxListEntity
+from pyqalx.core.entities.entity import AggregationResult
 from pyqalx.core.errors import (
     QalxError,
     QalxNoGUIDError,
@@ -17,8 +17,6 @@ from pyqalx.core.errors import (
     QalxFileError,
 )
 
-logger = logging.getLogger(__name__)
-
 
 class QalxAdapter(object):
     """
@@ -84,6 +82,15 @@ class QalxAdapter(object):
             self._entity_class.entity_type
         ]
 
+    @staticmethod
+    def raise_api_response_error(data_dict):
+        message = (
+            "API request error, message:\n\n-vvv-\n\n"
+            + "\n".join([f"{k}: {v}" for k, v in data_dict.items()])
+            + "\n\n-^^^-"
+        )
+        raise QalxAPIResponseError(message)
+
     def _process_api_request(self, method, *args, **kwargs):
         """calls to transport layer
 
@@ -92,11 +99,10 @@ class QalxAdapter(object):
         :param kwargs: kwargs to be passed to endpoint method
         :returns: `dict` containing API resource data
         """
-
         headers = kwargs.pop("headers", {})
 
         try:
-            json.dumps(kwargs)
+            bson_extra.dumps(kwargs)
         except (TypeError, ValueError, OverflowError):
             raise QalxError(
                 "One of the keyword arguments is not JSON serializable"
@@ -104,18 +110,15 @@ class QalxAdapter(object):
 
         try:
             endpoint = getattr(self.session.rest_api, method.lower())
-            logger.debug(str(endpoint))
+            self.session.log.debug(str(endpoint))
         except AttributeError:
             raise QalxError(f"{method} not recognised.")
 
-        success, data = endpoint(*args, json=kwargs, headers=headers)
+        success, data = endpoint(*args, data=kwargs, headers=headers)
 
         if success:
             return data
-        m = "API request error, message:\n\n-vvv-\n\n"
-        m += "\n".join([f"{k}: {v}" for k, v in data.items()])
-        m += "\n\n-^^^-"
-        raise QalxAPIResponseError(m)
+        self.raise_api_response_error(data)
 
     def detail_endpoint(self, guid, *args, **kwargs):
         """
@@ -154,13 +157,30 @@ class QalxAdapter(object):
             raise QalxError("`meta` kwarg must be a `dict`")
         return kwargs
 
+    def _process_api_response(self, response, entity_class=None, **kwargs):
+        """
+        Hook for processing the response from the API.  By default
+        returns an instance of self.entity_class.  kwargs can be provided
+        which will be passed through to entity_class
+        :param response: The response dict from the API
+        :param entity_class:The entity class that should be instantiated.
+        :return: An instance of `entity_class`
+        """
+        if entity_class is None:
+            entity_class = self.entity_class
+        return entity_class(response, **kwargs)
+
     def add(self, **kwargs):
         """
-        Adds a new instance of `self.entity_class`.  Provide valid `kwargs`
-        for the object you are trying to create
+        Adds a new instance of `self.entity_class` or a parent class that is
+        provided in `kwargs`.  Provide valid `kwargs` for the object you are
+        trying to create
 
-        :return: An instance of `self.entity_class`
+        :return:    An instance of `self.entity_class` or an instance of
+                    `parent_class` if this is provided in `kwargs`
         """
+        # If parent_class is provided, we return an instance of the parent class
+        parent_class = kwargs.pop("parent_class", None)
         if kwargs.pop("_is_validated", False) is False:
             # We may have already validated at this point if we are
             # doing a bulk insert.  This kwarg should be set on the adapter
@@ -180,8 +200,13 @@ class QalxAdapter(object):
         response = self._process_api_request(
             "post", self.list_endpoint(**kwargs), **kwargs
         )
-        # TODO: #3608 Should we return packable entiies in an unpacked state?
-        return self.entity_class(response)
+        if parent_class:
+            return self._process_api_response(
+                entity_class=parent_class,
+                child=self.entity_class,
+                response=response,
+            )
+        return self._process_api_response(response)
 
     def get(self, guid, *args, **kwargs):
         """
@@ -192,11 +217,11 @@ class QalxAdapter(object):
         """
         endpoint = self.detail_endpoint(guid=guid, *args, **kwargs)
         resp = self._process_api_request("get", endpoint, *args, **kwargs)
-        logger.debug(
+        self.session.log.debug(
             f"get {self.entity_class.entity_type} with "
             f"guid {guid} with {endpoint}"
         )
-        return self.entity_class(resp)
+        return self._process_api_response(resp)
 
     def find_one(self, query, **kwargs):
         """
@@ -212,12 +237,22 @@ class QalxAdapter(object):
         return self.find(many=False, limit=2, query=query, **kwargs)
 
     def find(
-        self, query=None, sort=None, skip=0, limit=25, many=True, **kwargs
+        self,
+        query=None,
+        aggregate=None,
+        sort=None,
+        skip=0,
+        limit=25,
+        many=True,
+        **kwargs,
     ):
         """
         Method for listing entities.
 
-        :param query: The optioanl Mongo query to find entities
+        :param query: The optional Mongo query to find entities
+        :param aggregate:   Optional Mongo aggregation pipeline argument that
+                            can be used to perform and return aggregation
+                            results
         :param sort: The keys to sort by
         :param skip: The number of results to skip (offset) by
         :param limit: How many results should the response be limited to
@@ -227,7 +262,7 @@ class QalxAdapter(object):
         """
         list_endpoint = self.list_endpoint(**kwargs)
 
-        logger.debug(
+        self.session.log.debug(
             f"find {self.entity_class.entity_type} with {list_endpoint}"
         )
 
@@ -235,7 +270,15 @@ class QalxAdapter(object):
             # We need to dump the query to JSON as there could be
             # complex lookups which would get lost when the request is made
             try:
-                query = json.dumps(query)
+                query = bson_extra.dumps(query)
+            except (TypeError, ValueError, OverflowError):
+                raise QalxError(
+                    "One of the keyword arguments is not JSON serializable"
+                )
+
+        if aggregate:
+            try:
+                aggregate = bson_extra.dumps(aggregate)
             except (TypeError, ValueError, OverflowError):
                 raise QalxError(
                     "One of the keyword arguments is not JSON serializable"
@@ -248,9 +291,16 @@ class QalxAdapter(object):
             skip=skip,
             limit=limit,
             query=query,
+            aggregate=aggregate,
             **kwargs,
         )
-        entities = QalxListEntity(resp, child=self.entity_class)
+        entity_class = AggregationResult if aggregate else QalxListEntity
+        entities = self._process_api_response(
+            resp, entity_class=entity_class, child=self.entity_class
+        )
+        # `many` argument is redundant if `aggregate` is supplied
+        if aggregate:
+            return entities
 
         if many is False:
 
@@ -303,13 +353,13 @@ class QalxAdapter(object):
 
         endpoint = self.detail_endpoint(guid=guid, *args, **kwargs)
         keys_to_save = self.get_keys_to_save(entity=entity)
-        logger.debug(
+        self.session.log.debug(
             f"save {self.entity_class.entity_type} with "
-            "guid {guid} with {endpoint}"
+            f"guid {guid} with {endpoint}"
         )
 
         resp = self._process_api_request("patch", endpoint, **keys_to_save)
-        return self.entity_class(resp)
+        return self._process_api_response(resp)
 
     def archive(self, entity, *args, **kwargs):
         """
@@ -320,13 +370,23 @@ class QalxAdapter(object):
         guid = entity["guid"]
         detail_endpoint = self.detail_endpoint(guid=guid, *args, **kwargs)
         endpoint = f"{detail_endpoint}/archive"
-        logger.debug(
+        self.session.log.debug(
             f"archive {self.entity_class.entity_type} with "
-            "guid {guid} with {endpoint}"
+            f"guid {guid} with {endpoint}"
         )
 
         resp = self._process_api_request("patch", endpoint)
-        return self.entity_class(resp)
+        return self._process_api_response(resp)
+
+    def delete(self, entity, *args, **kwargs):
+        """
+        Deletes the given entity from the API
+        :param entity: An instance of QalxEntity
+        """
+        detail_endpoint = self.detail_endpoint(
+            guid=entity["guid"], *args, **kwargs
+        )
+        return self._process_api_request("delete", detail_endpoint)
 
 
 class QalxUnpackableAdapter(QalxAdapter):
@@ -339,14 +399,14 @@ class QalxUnpackableAdapter(QalxAdapter):
         """
         The key that the child entities live in
         """
-        return self.child_entity_class.entity_type + "s"
+        return self.child_entity_class.pluralise()
 
     @staticmethod
     def _child_list_response(child_adapter, query, **kwargs):
-        return child_adapter.find(query=query)
+        return child_adapter.find(query=query, **kwargs)
 
     def _child_list_request(
-        self, entity, kids, child_adapter, query_key="guid"
+        self, entity, kids, child_adapter, query_key="guid", child_fields=None
     ):
         """
         Builds the list request for getting any unpackable children. Handles
@@ -358,24 +418,31 @@ class QalxUnpackableAdapter(QalxAdapter):
         :return: A list of unpacked child_adapter.entity_class entities
         """
         unpacked_entities = []
+
         for _page in entity._chunks(kids, chunk_size=100):
             # Chunk to avoid hitting maximum request size
             values = list(filter(None, _page))
             query = {query_key: {"$in": values}}
             _resp = self._child_list_response(
-                child_adapter=child_adapter, query=query, entity=entity
+                child_adapter=child_adapter, query=query, fields=child_fields
             )
             unpacked_entities += _resp["data"]
             while _resp["query"]["next"] is not None:
                 # We got a paginated response so keep going through the pages
                 # until we exhaust this chunk of data
-                logger.debug(
+                self.session.log.debug(
                     f"get {self.entity_class.entity_type} with "
-                    "{_resp['query']['next']}"
+                    f"{_resp['query']['next']}"
                 )
-                _resp = self._process_api_request("get", _resp["query"]["next"])
-                unpacked_entities += QalxListEntity(
-                    _resp, child=self.child_entity_class
+
+                _resp = self._process_api_request(
+                    "get", _resp["query"]["next"], fields=child_fields
+                )
+
+                unpacked_entities += self._process_api_response(
+                    _resp,
+                    entity_class=QalxListEntity,
+                    child=self.child_entity_class,
                 )["data"]
         return unpacked_entities
 
@@ -388,7 +455,7 @@ class QalxUnpackableAdapter(QalxAdapter):
         """
         return list(entity[self.kids].values())
 
-    def _get_child_entities(self, entity):
+    def _get_child_entities(self, entity, **kwargs):
         """
         Gets all the child entities for the given entity.  Handles chunking
         the children up into blocks to guard against sending a request that is
@@ -405,46 +472,63 @@ class QalxUnpackableAdapter(QalxAdapter):
             self.session, self.child_entity_class.entity_type
         )
         unpacked_entities = self._child_list_request(
-            entity=entity, kids=kids, child_adapter=child_adapter
+            entity=entity, kids=kids, child_adapter=child_adapter, **kwargs
         )
         return unpacked_entities
 
-    def _unpacked_entities_to_valid_children(self, entity, unpacked_entities):
+    def _unpacked_entities_to_valid_children(
+        self, entity, unpacked_entities, **kwargs
+    ):
         """
         Converts a list of `unpacked_entities` into valid child entities
-        for the given entity.
+        for the given entity.  Handles getting any nested children (i.e. items
+        on sets when we are unpacking a group).  A two stage process to keep
+        queries down in the instance where the same set is on a group multiple
+        times
 
         :param entity: The entity that has kids that should be unpacked
         :param unpacked_entities: The unpacked child entities that need to be
                                   stitched into the entity
-        :return: The unpacked entities in the correct format to be assigne to
+        :return: The unpacked entities in the correct format to be assigned to
                  `entity[self.kids]`
         """
         to_return = {}
-        for key, item_guid in entity[self.kids].items():
-            for child_entity in unpacked_entities:
-                if item_guid == child_entity["guid"]:
-                    child_entity_adapter = getattr(
-                        self.session, child_entity.entity_type
-                    )
-                    if hasattr(child_entity_adapter, "child_entity_class"):
-                        # The child entity has a child adapter of it's own.
-                        # Attempt unpack. This can occur if we are unpacking
-                        # a Group as we may need to unpack the Sets on the
-                        # Group
-                        to_return[key] = child_entity_adapter._attempt_unpack(
-                            child_entity
-                        )
-                    else:
-                        to_return[key] = child_entity
+        fully_unpacked_entities = []
+        for child_entity in unpacked_entities:
+            # iterate through all the unpacked_entities, getting any further
+            # nested children as required
+            child_entity_adapter = getattr(
+                self.session, child_entity.entity_type
+            )
+            if hasattr(child_entity_adapter, "child_entity_class"):
+                # The child entity has a child adapter of it's own.
+                # Attempt unpack. This can occur if we are unpacking
+                # a Group as we may need to unpack the Sets on the
+                # Group
+                child_entity_with_nested = child_entity_adapter._attempt_unpack(
+                    child_entity, **kwargs
+                )
+                # At this point this given `child_entity` will have all
+                # the nested children within fully unpacked.
+                fully_unpacked_entities.append(child_entity_with_nested)
+            else:
+                # This is the bottom of the unpacking hierarchy.
+                fully_unpacked_entities.append(child_entity)
+
+        for key, child_guid in entity[self.kids].items():
+            # Iterate through the key/values on the children of this entity
+            # and update the appropriate key from the fully_unpacked_entities.
+            for child_entity in fully_unpacked_entities:
+                if child_guid == child_entity["guid"]:
+                    to_return[key] = child_entity
         return to_return
 
     def _should_unpack(self):
-        return self.session.config.getboolean(
+        return self.session.config[
             "UNPACK_" + self.entity_class.entity_type.upper()
-        )
+        ]
 
-    def _attempt_unpack(self, entity):
+    def _attempt_unpack(self, entity, **kwargs):
         """
         For the given entity will attempt
         to unpack the child_entity data that it contains
@@ -456,9 +540,9 @@ class QalxUnpackableAdapter(QalxAdapter):
         if self._should_unpack() and self.kids in entity:
             # `self.kid` might not be there if we have specified a subset
             # of fields
-            unpacked_entities = self._get_child_entities(entity)
+            unpacked_entities = self._get_child_entities(entity, **kwargs)
             entity[self.kids] = self._unpacked_entities_to_valid_children(
-                entity=entity, unpacked_entities=unpacked_entities
+                entity=entity, unpacked_entities=unpacked_entities, **kwargs
             )
         return entity
 
@@ -515,26 +599,38 @@ class QalxUnpackableAdapter(QalxAdapter):
         """
         return {k: v for k, v in kids.items()}
 
-    def get(self, guid, *args, **kwargs):
+    def get(self, guid, child_fields=None, *args, **kwargs):
         """
         Unpackable entities need to be unpacked after retrieval
 
         :param guid: The `guid` of the entity to get
+        :param child_fields: A list of fields that should be returned from
+        child entities
+        :type child_fields: list
         :return: An unpacked entity
         """
         entity = super(QalxUnpackableAdapter, self).get(
             guid=guid, *args, **kwargs
         )
-        entity = self._attempt_unpack(entity)
+        entity = self._attempt_unpack(entity, child_fields=child_fields)
         return entity
 
-    def find(self, many=True, *args, **kwargs):
+    def find(self, many=True, child_fields=None, *args, **kwargs):
+        """
+        Return multiple packed entities from the API
+        :param many:Are many results expected or just one?
+        :type many:bool
+        :param child_fields: A list of fields that should be returned from
+        child entities
+        :type child_fields: list
+        :return:A list of entities
+        """
         entities = super(QalxUnpackableAdapter, self).find(
             many=many, *args, **kwargs
         )
         if many is False:
             # Only unpack a list view if we have a single entity
-            entities = self._attempt_unpack(entities)
+            entities = self._attempt_unpack(entities, child_fields=child_fields)
         return entities
 
     def save(self, entity, *args, **kwargs):
@@ -556,13 +652,24 @@ class QalxUnpackableAdapter(QalxAdapter):
             entity = self._attempt_unpack(entity=entity)
         return entity
 
+    def cache_kids(self, kwargs):
+        return kwargs.get(self.kids, None)
+
+    def save_kids_on_entity(self, entity, kids):
+        entity[self.kids] = kids
+
     def add(self, **kwargs):
         # Need to validate before we pack the kids up
         kwargs = self._validate(**kwargs)
+        # Cache the kids to reassign them to the newly created entity
+        cached_kids = self.cache_kids(kwargs)
         kwargs = self._pack_kids_for_request(kwargs)
-        return super(QalxUnpackableAdapter, self).add(
+        entity = super(QalxUnpackableAdapter, self).add(
             _is_validated=True, **kwargs
         )
+        if cached_kids is not None:
+            self.save_kids_on_entity(entity, cached_kids)
+        return entity
 
 
 class QalxSignalAdapter(QalxAdapter):
@@ -588,7 +695,7 @@ class QalxSignalAdapter(QalxAdapter):
         """
         guid = entity["guid"]
         endpoint = self.detail_endpoint(guid=guid, **kwargs)
-        logger.debug(
+        self.session.log.debug(
             f"terminate {self.entity_class.entity_type} with guid {guid} with {endpoint}"  # noqa
         )
         signal = self.signal_class._terminate_signal(*args, **kwargs)
@@ -599,7 +706,7 @@ class QalxSignalAdapter(QalxAdapter):
         guid = entity["guid"]
         endpoint = self.detail_endpoint(guid=guid, **kwargs)
         signal = self.signal_class._stop_signal(stop)
-        logger.debug(
+        self.session.log.debug(
             f"signal {signal} {self.entity_class.entity_type} with guid {guid} with {endpoint}"  # noqa
         )
         self._process_api_request("patch", endpoint, signal=signal)
@@ -729,18 +836,31 @@ class QalxFileAdapter(QalxAdapter):
     Adapter that allows files to be uploaded to an entity
     """
 
-    def _validate(self, input_file, file_name, **kwargs):
+    def _process_api_response(self, *args, **kwargs):
         """
-        Ensures that the `input_file` and/or `file_name` are present
+        Ensures that if there is a KEYFILE configured that this is present
+        on all instances of `self.entity_class` to provide seamless decryption
+        of files when downloading
+        :param response: The response dict from the API
+        :return: An instance of `self.entity_class`
+        """
+        kwargs["keyfile"] = self.session.config.get("KEYFILE")
+        return super(QalxFileAdapter, self)._process_api_response(
+            *args, **kwargs
+        )
+
+    def _validate(self, source, file_name, **kwargs):
+        """
+        Ensures that the `source` and/or `file_name` are present
         and correct
-        :param input_file: The path to the file or the file stream
+        :param source: The path to the file or the file stream
         :param file_name: The name of the file
         :return: kwargs
         """
-        if input_file is not None:
+        if source is not None:
             # Only validate the file is it's specified in the
             # implementing adapter
-            is_filestream = self.session.rest_api.is_filestream(input_file)
+            is_filestream = self.session.rest_api.is_filestream(source)
             if is_filestream:
                 if not file_name:
                     raise QalxFileError(
@@ -749,56 +869,73 @@ class QalxFileAdapter(QalxAdapter):
                     )
                 file = {"name": file_name}
             else:
-                if os.path.isfile(input_file):
-                    file_path = os.path.abspath(input_file)
+                if os.path.isfile(source):
+                    file_path = os.path.abspath(source)
                     filename = (
                         file_name if file_name else os.path.basename(file_path)
                     )
                     file = {"name": filename}
                 else:
-                    raise QalxFileError(
-                        "Invalid file `{0}`".format(input_file)
-                    )  # noqa
+                    raise QalxFileError(f"Invalid file `{source}`")
+            key_file = self.session.config.get("KEYFILE")
+            if key_file is not None:
+                # The user can suppress encryption on a per-case basis, by
+                # passing `encrypt=False`
+                if kwargs["encrypt"]:
+                    file["keyfile"] = os.path.basename(key_file)
             kwargs[self.entity_class._file_key] = file
         return super(QalxFileAdapter, self)._validate(**kwargs)
 
-    def add(self, input_file, file_name, upload=True, **kwargs):
+    def add(self, source, file_name, upload=True, encrypt=True, **kwargs):
         """
         When a user creates an entity with a file pyqalx will upload the file
         automatically for you after the entity itself has been created.  Pass
         `upload=False` through to stop this from happening and upload the file
         yourself
-        :param input_file: The input file path or stream
+        :param source: The input file path or stream
         :param file_name: The optional file name
         :param upload: Whether the file should be automatically uploaded or not
+        :param encrypt: Whether the file should be automatically encrypted if
+         the KEYFILE is present
         :return: A valid entity
         """
         entity = super(QalxFileAdapter, self).add(
-            input_file=input_file, file_name=file_name, **kwargs
+            source=source, file_name=file_name, encrypt=encrypt, **kwargs
         )
-        if upload and input_file is not None:
-            entity = self._upload_file(entity=entity, input_file=input_file)
+        if upload and source is not None:
+            entity = self._upload_file(entity=entity, source=source)
         return entity
 
     def save(
-        self, entity, input_file, file_name, upload=True, *args, **kwargs
+        self,
+        entity,
+        source,
+        file_name,
+        upload=True,
+        encrypt=True,
+        *args,
+        **kwargs,
     ):  # noqa
         """
         Saves an entity - optionally updating any file data.
         :param entity: The entity that is being saved
-        :param input_file: The input file path or stream
+        :param source: The input file path or stream
         :param file_name: The optional file name
         :param upload: Whether the file should be automatically uploaded or not
+        :param encrypt: Whether the file should be automatically encrypted if
+         the KEYFILE is present
         :return: A valid entity
         """
         entity = super(QalxFileAdapter, self).save(
-            entity=entity, input_file=input_file, file_name=file_name, **kwargs
+            entity=entity,
+            source=source,
+            file_name=file_name,
+            encrypt=encrypt,
+            **kwargs,
         )
-        if input_file is not None:
+        if source is not None:
             if upload:
-                entity = self._upload_file(
-                    entity=entity, input_file=input_file
-                )  # noqa
+                entity = self._upload_file(entity=entity, source=source)  # noqa
             else:
                 # Entity is returned with the put_url so that the user can
                 # upload the file manually
@@ -807,21 +944,18 @@ class QalxFileAdapter(QalxAdapter):
         entity[self.entity_class._file_key].pop("put_url", None)
         return entity
 
-    def _upload_file(self, entity, input_file):
+    def _upload_file(self, entity, source):
         # Only upload if everything was OK with the item creation
         # and the user wants to upload the file automatically.
+
         s3_resp_ok, s3_data = self.session.rest_api._upload_to_s3(
-            entity, input_file, self.entity_class._file_key
-        )  # noqa
+            entity, source, self.entity_class._file_key
+        )
         if s3_resp_ok is False:
             # The S3 upload failed, return the failed S3 response in place
             # of the response from the api.  Any implementing client will
             # have to handle tidying up any incorrect data.
-            # TODO: #3928 Handle errors in a consistent way in adapters
-            m = "API request error, message:\n\n-vvv-\n\n"
-            m += "\n".join([f"{k}: {v}" for k, v in s3_data.items()])
-            m += "\n\n-^^^-"
-            raise QalxAPIResponseError(m)
+            self.raise_api_response_error(s3_data)
         else:
             # Always pop the `put_url` off if it's a success
             entity[self.entity_class._file_key].pop("put_url")
@@ -840,7 +974,7 @@ class QalxUpdateStatusAdapter(QalxAdapter):
         """
         guid = entity["guid"]
         endpoint = self.detail_endpoint(guid=guid, **kwargs)
-        logger.debug(
+        self.session.log.debug(
             f"update status{self.entity_class.entity_type} with guid {guid} with {endpoint}"  # noqa
         )
         status_dict = {"message": status}
