@@ -1,10 +1,10 @@
 from pyqalx.core.adapters.adapter import QalxUnpackableAdapter
-from pyqalx.core.entities import Group, Set
+from pyqalx.core.entities import Group
 
 
 class QalxGroup(QalxUnpackableAdapter):
     _entity_class = Group
-    child_entity_class = Set
+    child_entity_class = Group.child_entity_class
 
     def add(self, sets, meta=None, **kwargs):
         """
