@@ -1,17 +1,12 @@
-import deprecation
-import logging
+import concurrent
 
-from pyqalx.__version__ import __version__
 from pyqalx.core.adapters.adapter import (
     QalxValidateBlueprintAdapter,
     QalxFileAdapter,
 )
 from pyqalx.core.entities import Item
 from pyqalx.core.entities.item import ItemAddManyEntity
-from pyqalx.core.errors import QalxFileError, QalxError
-
-
-logger = logging.getLogger(__name__)
+from pyqalx.core.errors import QalxError, QalxAPIResponseError
 
 
 class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
@@ -20,7 +15,7 @@ class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
     def _validate(
         self,
         data=None,
-        input_file=None,
+        source=None,
         file_name="",
         meta=None,
         blueprint_name=None,
@@ -30,7 +25,7 @@ class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
             data = {}
         return super(QalxItem, self)._validate(
             data=data,
-            input_file=input_file,
+            source=source,
             file_name=file_name,
             meta=meta,
             blueprint_name=blueprint_name,
@@ -40,7 +35,7 @@ class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
     def add(
         self,
         data=None,
-        input_file=None,
+        source=None,
         file_name="",
         meta=None,
         blueprint_name=None,
@@ -52,11 +47,11 @@ class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
 
         :param data: Optional data to store against this Item
         :type data: dict
-        :param input_file: file path or instance of
+        :param source: file path or instance of
             StringIO or BytesIO (https://docs.python.org/3/library/io.html)
-        :type input_file: str or `io` object
+        :type source: str or `io` object
         :param file_name: input file name. Optional if a file path is given
-         for input_file.  Required if an `io` object
+         for source.  Required if an `io` object
         :type file_name: str
         :param meta: A dictionary of metadata to store
         :type meta: dict
@@ -67,7 +62,7 @@ class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
 
         return super(QalxItem, self).add(
             data=data,
-            input_file=input_file,
+            source=source,
             file_name=file_name,
             meta=meta,
             blueprint_name=blueprint_name,
@@ -75,12 +70,7 @@ class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
         )
 
     def save(
-        self,
-        entity,
-        input_file=None,
-        file_name="",
-        blueprint_name=None,
-        **kwargs,
+        self, entity, source=None, file_name="", blueprint_name=None, **kwargs
     ):
         """
         Saves an updated existing `Item` instance.
@@ -89,11 +79,11 @@ class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
 
         :param entity: The entity that we are saving
         :type entity: An instance of `Item`
-        :param input_file: file path or instance of StringIO or
+        :param source: file path or instance of StringIO or
             BytesIO (https://docs.python.org/3/library/io.html)
-        :type input_file: str or `io` object
+        :type source: str or `io` object
         :param file_name: input file name. Optional if a file path is given
-            for input_file.  Required if an `io` object
+            for source.  Required if an `io` object
         :type file_name: str
         :type file_name: str
         :param meta: A dictionary of metadata to store
@@ -106,13 +96,13 @@ class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
         """
         return super(QalxItem, self).save(
             entity=entity,
-            input_file=input_file,
+            source=source,
             file_name=file_name,
             blueprint_name=blueprint_name,
             **kwargs,
         )
 
-    def add_many(self, items, **kwargs):
+    def add_many(self, items, encrypt=True, **kwargs):
         """
         Adds multiple items at once.
         :param items: A list of item data that you want to create.  This can
@@ -120,6 +110,10 @@ class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
         the list should be the same as if a single item were being created via
         the `add` method
         :type items: list
+        :param encrypt: Whether the file should be automatically encrypted if
+        the KEYFILE is present.  All files are encrypted or not encrypted based
+        on this value
+        :type encrypt: bool
         :return:
         """
         if not isinstance(items, list):
@@ -131,7 +125,7 @@ class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
             # Ensure all the items are valid.  They should be in the same
             # format as if a user is adding a single item.  This also handles
             # checking the blueprint for each item
-            validated_items.append(self._validate(**item))
+            validated_items.append(self._validate(encrypt=encrypt, **item))
 
         resp = self.add(
             items=validated_items,
@@ -140,6 +134,7 @@ class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
             # Don't upload as we need to handle
             # the file upload manually
             upload=False,
+            parent_class=ItemAddManyEntity,
             **kwargs,
         )
 
@@ -151,16 +146,38 @@ class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
         ERRORED_KEY = None
         for _ in resp["errors"].keys():
             resp["items"].insert(int(_), ERRORED_KEY)
+        uploaded_files = {}
+
+        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
+            # Allows for partial success (i.e. some files uploaded successfully,
+            # others did not)
+            for index, item in enumerate(items):
+                # Iterate the original `items` provided by the user
+                created_item = resp["items"][index]
+                if item.get("source") and created_item is not None:
+                    # There was a file on the original item and it was also
+                    # successfully created by the API.  Upload the file as normal
+                    uploaded_files[
+                        executor.submit(
+                            self._upload_file,
+                            entity=resp["items"][index],
+                            source=item["source"],
+                        )
+                    ] = index
+            for future in concurrent.futures.as_completed(uploaded_files):
+                # Handles updating the error response if any files failed to
+                # upload
+                index = uploaded_files[future]
+                try:
+                    future.result()
+                except QalxAPIResponseError as exc:
+                    # There was an error uploading this single file.  Update
+                    # the response dict accordingly.
+                    resp["items"].pop(index)
+                    resp["items_created"] = resp["items_created"] - 1
+                    resp["items_errored"] = resp["items_errored"] + 1
+                    resp["errors"][index] = exc
 
-        for index, item in enumerate(items):
-            # Iterate the original `items` provided by the user
-            created_item = resp["items"][index]
-            if item.get("input_file") and created_item is not None:
-                # There was a file on the original item and it was also
-                # successfully created by the API.  Upload the file as normal
-                self._upload_file(
-                    entity=resp["items"][index], input_file=item["input_file"]
-                )
         # Finally, remove all the `None` values from the items as these were
         # errors that we are done with
         resp["items"] = list(
@@ -168,57 +185,9 @@ class QalxItem(QalxValidateBlueprintAdapter, QalxFileAdapter):
         )
         if resp["errors"]:
             for index, error in resp["errors"].items():
-                logger.warning(
+                self.session.log.warning(
                     f"Error bulk creating at index `{index}`. "
                     f"Error: `{error}`. "
                     f"Original Data: `{items[int(index)]}`"
                 )
-        return ItemAddManyEntity(resp, child=self.entity_class)
-
-    @deprecation.deprecated(
-        deprecated_in="0.6",
-        removed_in="1.0",
-        current_version=__version__,
-        details="Use the `add` function instead",
-    )
-    def add_file(
-        self,
-        input_file,
-        data=None,
-        meta=None,
-        file_name="",
-        blueprint_name=None,
-        **kwargs,
-    ):
-        """adds a file to qalx with optional data and metadata
-
-        :param input_file: input file path or stream
-        :type input_file: str or `filestream`
-        :param data: optional data you want to store with this file
-        :type data: dict
-        :param meta: additional data about the item
-        :param meta: dict
-        :param file_name: input file name. Optional for file path. Required
-                          for file stream
-        :type file_name: str
-        :param blueprint_name: An optional blueprint name to use if you want
-                              to validate this item against an existing
-                              Blueprint
-        :type blueprint_name: str
-        :return: pyqalx.core.Item
-        """
-        if data is None:
-            data = {}
-
-        if self.session.rest_api.is_filestream(input_file) and not file_name:
-            raise QalxFileError(
-                "You must supply a file name when supplying a file stream"
-            )
-        return self.add(
-            input_file=input_file,
-            file_name=file_name,
-            data=data,
-            meta=meta,
-            blueprint_name=blueprint_name,
-            **kwargs,
-        )
+        return resp
