@@ -12,9 +12,7 @@ class QalxQueue(QalxNamedEntityAdapter):
         The configurable parameters for a `Queue`
         :return:
         """
-        return {
-            "VisibilityTimeout": int(self.session.config["MSG_BLACKOUTSECONDS"])
-        }
+        return {"VisibilityTimeout": self.session.config["MSG_BLACKOUTSECONDS"]}
 
     def add(self, name, meta=None, **kwargs):
         """
@@ -41,9 +39,9 @@ class QalxQueue(QalxNamedEntityAdapter):
         :return: A list of `QueueMessage` instances
         """
         config = self.session.config
-        max_msgs = int(config["Q_MSGBATCHSIZE"])
-        visibility = int(config["MSG_BLACKOUTSECONDS"])
-        waittime = int(config["MSG_WAITTIMESECONDS"])
+        max_msgs = config["Q_MSGBATCHSIZE"]
+        visibility = config["MSG_BLACKOUTSECONDS"]
+        waittime = config["MSG_WAITTIMESECONDS"]
 
         message = worker.queue.get_messages(
             max_num_msg=max_msgs,
