@@ -2,14 +2,12 @@ from pyqalx.core.adapters.adapter import (
     QalxUnpackableAdapter,
     QalxValidateBlueprintAdapter,
 )
-from pyqalx.core.entities import Item, Set
+from pyqalx.core.entities import Set
 
 
-class QalxSet(
-    QalxValidateBlueprintAdapter, QalxUnpackableAdapter,
-):
+class QalxSet(QalxValidateBlueprintAdapter, QalxUnpackableAdapter):
     _entity_class = Set
-    child_entity_class = Item
+    child_entity_class = Set.child_entity_class
 
     def add(self, items, meta=None, blueprint_name=None, **kwargs):
         """
