@@ -3,12 +3,14 @@ from itertools import zip_longest
 
 import requests
 
+from pyqalx.core.encryption import QalxEncrypt
 from pyqalx.core.entities.object_dict import ObjectDict
 
 from pyqalx.core.errors import (
     QalxAPIResponseError,
     QalxEntityTypeNotFound,
     QalxError,
+    QalxQueueError,
 )
 
 
@@ -58,12 +60,17 @@ class QalxListEntity(ObjectDict):
                 " it to be a list".format(self._data_key)
             )
         # Cast all the entities in data to be an instance of `self.child`
+        keyfile = kwargs.pop("keyfile", None)
         self[self._data_key] = [
-            self.child(e) for e in pyqalxapi_list_response_dict[self._data_key]
-        ]  # noqa
+            self.child(e, keyfile=keyfile)
+            for e in pyqalxapi_list_response_dict[self._data_key]
+        ]
 
-    def __str__(self):
-        return f"[{self.child.entity_type} list]"
+
+class AggregationResult(QalxListEntity):
+    """
+    Class for aggregation result responses from the API
+    """
 
 
 class QalxEntity(ObjectDict):
@@ -94,12 +101,9 @@ class QalxEntity(ObjectDict):
 
     entity_type: str
 
-    def __init__(self, pyqalxapi_dict):
+    def __init__(self, pyqalxapi_dict, *args, **kwargs):
         super().__init__(pyqalxapi_dict)
 
-    def __str__(self):
-        return f"[{self.entity_type}] {self['guid']}"
-
     @classmethod
     def _chunks(cls, _iterable, chunk_size, fillvalue=None):
         """
@@ -110,6 +114,19 @@ class QalxEntity(ObjectDict):
         args = [iter(_iterable)] * chunk_size
         return zip_longest(fillvalue=fillvalue, *args)
 
+    @classmethod
+    def pluralise(cls):
+        """
+        Pluralises the entity type
+        """
+        vowels = {"a", "e", "i", "o", "u", "y"}
+        singular = cls.entity_type
+        plural = singular + "s"
+        if len(singular) > 1 and singular.endswith("y"):
+            if singular[-2] not in vowels:
+                plural = singular[:-1] + "ies"
+        return plural
+
     def __super_setattr__(self, name, value):
         """
         Convenience method for setting proper attributes on the entity class.
@@ -135,13 +152,26 @@ class QalxFileEntity(QalxEntity):
     _file_bytes = None
     _file_key = "file"
 
+    def __init__(self, *args, **kwargs):
+        """
+        QalxFileEntity subclasses can be provided with a `keyfile` kwarg.
+        This gets set on the instance of the entity and will allow for seamless
+        decryption of files
+        """
+        self.__super_setattr__("_key_file", kwargs.pop("keyfile", None))
+        super(QalxFileEntity, self).__init__(*args, **kwargs)
+
     def read_file(self):
         """
         If this Item contains a file, will read the file data and cache it
         against the Item.
 
-        :return: The content of the URL as a bytes object.  Accessible from
-            the `_file_bytes` attribute
+        :param key_file:    The path to the encryption key to be used for
+                            decrypting the file data. If it is not provided
+                            and the data was encrypted, the data will be read
+                            in the encrypted format
+        :return:    The content of the URL as a bytes object.  Accessible from
+                    the `_file_bytes` attribute
         :raises: pyqalx.errors.QalxError
         """
         if not self.get(self._file_key):
@@ -149,7 +179,14 @@ class QalxFileEntity(QalxEntity):
         else:
             response = requests.get(url=self[self._file_key]["url"])
             if response.ok:
-                self.__super_setattr__("_file_bytes", response.content)
+                file_bytes = response.content
+                if (
+                    self[self._file_key].get("keyfile") is not None
+                    and self._key_file is not None
+                ):
+                    qalx_encrypt = QalxEncrypt(self._key_file)
+                    file_bytes = qalx_encrypt.decrypt(file_bytes)
+                self.__super_setattr__("_file_bytes", file_bytes)
                 return self._file_bytes
             else:
                 raise QalxError(
@@ -168,6 +205,11 @@ class QalxFileEntity(QalxEntity):
         :param filename: The optional name of this file. Defaults to the name
             of the file on the instance
         :type filename: str
+        :param key_file:    The path to the encryption key to be used for
+                            decrypting the file data. If it is not provided
+                            and the data was encrypted, the data will be read
+                            in the encrypted format
+        :type key_file:     str
         :raises: pyqalx.errors.QalxError
         """
         if filename is None:
@@ -178,3 +220,49 @@ class QalxFileEntity(QalxEntity):
         with open(_filepath, "wb") as f:
             f.write(self._file_bytes)
         return _filepath
+
+
+class QalxQueueableEntity(QalxEntity):
+    """
+    A mixin which allows an entity to be sumitted to queue
+    """
+
+    @classmethod
+    def add_to_queue(cls, payload, queue, children=False, **message_kwargs):
+        """
+        Submits an entity, entity guid or list of entities/entity guids
+        to the given queue
+        :param payload:subclassed instance of ~entities.entity.QalxEntity,
+        a guid, or a list containing a combination of both
+        :param queue: An instance of ~entities.queue.Queue
+        Usage:
+        :type queue:~entities.queue.Queue
+        :param children:Whether we are submitting the child entities to the
+        queue rather than the given entity
+        :type children: bool
+        Example usage for an Item
+        Item.add_to_queue(item, queue)
+        Item.add_to_queue(item.guid, queue)
+        Item.add_to_queue([item, item.guid], queue)
+        """
+        entity_type = cls.entity_type
+        if children:
+            #  Get the children off this entity and submit them to the queue
+            if not hasattr(cls, "child_entity_class"):
+                raise QalxQueueError(
+                    f"`{cls}` is not an entity that " f"has children."
+                )
+            if not isinstance(payload, cls):
+                raise QalxQueueError(
+                    f"Expected payload of type `{cls}"
+                    f"` when calling `{cls.__name__}.add_to_queue` "
+                    f"with `children=True`. "
+                    f"Got `{type(payload)}`."
+                )
+            # SQS needs the guid as a string - not UUID
+            message_kwargs[f"parent_{entity_type}_guid"] = str(payload["guid"])
+            entity_type = cls.child_entity_class.entity_type
+            payload = list(payload[cls.child_entity_class.pluralise()].values())
+        return queue._send_message(
+            payload=payload, entity_type=entity_type, **message_kwargs
+        )
