@@ -5,7 +5,6 @@ contained in these mix-ins keeps the main factory class cleaner
 """
 
 import concurrent.futures
-import json
 import logging
 import shutil
 import subprocess
@@ -19,6 +18,9 @@ import stat
 import yaml
 from yaml import YAMLError
 
+from bson_extra import bson_extra
+
+from pyqalx.config import UserConfig, BotConfig
 from pyqalx.factories.bot_sources import BotSource, QalxItemBotSource
 from pyqalx.core.entities.factory import Factory as FactoryEntity
 from pyqalx.core.errors import (
@@ -32,9 +34,7 @@ from pyqalx.factories.sectors.manager import SectorManager
 from pyqalx.factories.sectors.sectors import Sectors
 from pyqalx.factories.sectors.providers import _providers, LocalProvider
 from pyqalx.factories.tree import QalxTree
-from pyqalx.factories.utils import _msg_user
-
-logger = logging.getLogger(__name__)
+from pyqalx.core.utils import _msg_user
 
 
 class FactoryValidateMixin:
@@ -131,7 +131,9 @@ class FactoryPackMixin:
             BotSourceClass = BotSource.get_source(bot)
         except KeyError as exc:
             raise QalxFactoryPackError(exc)
-        bot_source = BotSourceClass(bot_name, bot, self.build_path)
+        bot_source = BotSourceClass(
+            bot_name, bot, self.build_path, log=self.session.log
+        )
         bot_source.download_bot()
         self.bot_sources.append(bot_source)
 
@@ -161,7 +163,9 @@ class FactoryPackMixin:
                     subprocess.CalledProcessError,
                     NotADirectoryError,
                     PermissionError,
+                    FileNotFoundError,
                     QalxFactoryPackError,
+                    QalxFactoryValidationError,
                 ) as exc:
                     # Provides a common way of handling all the potential errors
                     # that could occur when packing bots
@@ -170,7 +174,9 @@ class FactoryPackMixin:
                     # subprocess.CalledProcessError: There was an error pip
                     #                                downloading the package
                     # NotADirectoryError: The specified file wasn't a directory
+                    # FileNotFoundError: The local path could not be found
                     # PermissionError: No permission to read from directory
+                    # QalxFactoryValidationError: Error validating bot path
 
                     # Grab this exception and raise later.  Need to wait for
                     # all processing to finish before exiting otherwise it won't
@@ -224,7 +230,7 @@ class FactoryBuildMixin:
         }
         self.entity = self.session.factory.add(
             name=self._plan_name,
-            input_file=plan,
+            source=plan,
             stage=self.stage,
             meta=factory_meta,
         )
@@ -236,7 +242,7 @@ class FactoryBuildMixin:
         bot on a factory stage. For each bot and bot source, a qalx item is
         created with the following specification:
 
-            input_file: An input file is provided and it is the zipped
+            source: An input file is provided and it is the zipped
                         directory of the bot source code
             meta:       The metadata, with the following key/value pairs:
                 "_class":           "pyqalx.factory"
@@ -261,7 +267,7 @@ class FactoryBuildMixin:
                 zip_path = bot_source.zip_build_path()
                 if zip_path:
                     # pypi bots won't have a file
-                    item["input_file"] = zip_path
+                    item["source"] = zip_path
                 item_dict_list.append(item)
         self.bot_items = self.session.item.add_many(item_dict_list)["items"]
 
@@ -408,6 +414,7 @@ class FactoryBuildMixin:
                     bot=bot,
                     build_path_dir=self.build_path,
                     bot_code_item=bot_code_item,
+                    log=self.session.log,
                 )
                 results.append(executor.submit(bot_source.download_bot))
 
@@ -488,7 +495,7 @@ class FactoryBuildMixin:
             if not isinstance(value, str):
                 # This value isn't a string.  Dump to JSON so it can be saved it
                 # in the correct structure
-                value = json.dumps(value)
+                value = bson_extra.dumps(value)
             arguments += ["--meta", f"{key}={value}"]
 
         # subprocess will be called using the same environment as the main
@@ -509,7 +516,7 @@ class FactoryBuildMixin:
             [
                 qalx,
                 "--user-profile",
-                sector[Sectors.QALX_PROFILE_KEY],
+                sector[Sectors.USER_PROFILE_KEY],
                 "--bot-profile",
                 sector[Sectors.BOT_PROFILE_KEY],
                 "bot-start",
@@ -706,7 +713,7 @@ class FactoryDemolishMixin:
 
         :return bot_name
         """
-        _msg_user(msg=f"[{plan_bot_name}] terminating")
+        self._msg_user(msg=f"[{plan_bot_name}] terminating")
         bots = []
         for (_, _plan_bot_name), bot in self.bots.items():
             if _plan_bot_name == plan_bot_name:
@@ -716,7 +723,7 @@ class FactoryDemolishMixin:
             self.session.bot.terminate(entity=bot, warm=True)
         for bot in bots:
             self.wait_on_bot_termination(bot)
-        _msg_user(msg=f"[{plan_bot_name}] terminated successfully")
+        self._msg_user(msg=f"[{plan_bot_name}] terminated successfully")
         return plan_bot_name
 
     def concurrent_terminate(self, workflow, bot_names):
@@ -747,7 +754,7 @@ class FactoryDemolishMixin:
                     #                        Unlikely to happen if the user gets
                     #                        this far but they should be made
                     #                        aware so they can take manual steps
-                    _msg_user(
+                    self._msg_user(
                         msg=f"Exception occurred when "
                         f"terminating {bot_name}.  "
                         f"You may need to terminate this "
@@ -819,6 +826,9 @@ class BaseFactory(
         self.bot_items = []
         self._bots = None
 
+    def _msg_user(self, *args, **kwargs):
+        _msg_user(log=self.session.log, *args, **kwargs)
+
     @property
     def build_path(self):
         """
@@ -861,7 +871,7 @@ class BaseFactory(
                 # Caches the entity to avoid repeated lookups
                 return
             try:
-                uuid.UUID(factory)
+                uuid.UUID(str(factory))
             except ValueError:
                 raise QalxFactoryError(
                     f"factory not a valid {FactoryEntity} "
@@ -967,11 +977,11 @@ class BaseFactory(
 
     @property
     def _stage_user_profile(self):
-        return self._plan_stage.get("qalx_profile", "default")
+        return self._plan_stage.get("user_profile", UserConfig.default)
 
     @property
     def _stage_bot_profile(self):
-        return self._plan_stage.get("bot_profile", "default")
+        return self._plan_stage.get("bot_profile", BotConfig.default)
 
     @property
     def stacks(self):
@@ -1018,9 +1028,6 @@ class BaseFactory(
             bots = self.session.bot.find(query=query)
 
             for bot_entity in bots["data"]:
-                # TODO: id:3827 When the API returns workers in bots, this reload
-                #  won't be required
-                bot_entity = self.session.bot.reload(bot_entity)
                 plan_bot_name = bot_entity.meta[namespace_key][
                     plan_bot_name_key
                 ]
@@ -1034,7 +1041,9 @@ class BaseFactory(
         Deletes the entire build path from the filesystem if it exists. If the
         directory is not found, a message is printed out
         """
-        _msg_user(msg=f"Deleting data from build path: `{self.build_path}`")
+        self._msg_user(
+            msg=f"Deleting data from build path: `{self.build_path}`"
+        )
 
         def remove_readonly(func, path, _):
             """Clear the readonly bit and reattempt the removal"""
@@ -1044,7 +1053,7 @@ class BaseFactory(
         if os.path.exists(self.build_path):
             shutil.rmtree(self.build_path, onerror=remove_readonly)
         else:
-            _msg_user(
+            self._msg_user(
                 msg=f"`{self.build_path}` not found on this machine.  "
                 f"Unable to delete"
             )
@@ -1063,31 +1072,29 @@ class BaseFactory(
         }
         items = self.session.item.find(query=query)
         for item in items["data"]:
-            deleted_successfully, details = self.session.rest_api.delete(
-                f"item/{item['guid']}"
-            )
-            if not deleted_successfully:
+            try:
+                self.session.item.delete(item)
+            except QalxAPIResponseError as exc:
                 err_msg = (
                     f"There was an error deleting a qalx item containing "
                     f"factory bot code.  "
-                    f"You may have to delete this manually.: {details}"
+                    f"You may have to delete this manually.: {exc}"
                 )
-                _msg_user(msg=err_msg, level=logging.ERROR)
+                self._msg_user(msg=err_msg, level=logging.ERROR)
 
     def _delete_factory_build(self):
         """
         Deletes the factory build entity from the API. In the case of an error,
         the error message is printed and logged
         """
-        deleted_successfully, details = self.session.rest_api.delete(
-            f"factory/{self.entity.guid}"
-        )
-        if not deleted_successfully:
+        try:
+            self.session.factory.delete(self.entity)
+        except QalxAPIResponseError as exc:
             err_msg = (
                 f"There was an error deleting the factory build entity.  "
-                f"You may have to delete this manually.: {details}"
+                f"You may have to delete this manually.: {exc}"
             )
-            _msg_user(msg=err_msg, level=logging.ERROR)
+            self._msg_user(msg=err_msg, level=logging.ERROR)
 
     def _delete_stacks(self):
         """
