@@ -39,7 +39,8 @@ class AWSStack:
         """
         tags = [
             self._add_tag("Name", name),
-            self._add_tag("qalx-factory-guid", self.factory_entity.guid),
+            # This should be stored as a string on AWS not uuid.UUID
+            self._add_tag("qalx-factory-guid", str(self.factory_entity.guid)),
             self._add_tag("qalx-factory-stage", self.provider.stage),
             self._add_tag("qalx-created-version", __version__),
             self._add_tag("qalx-factory-name", self.factory_entity.name),
@@ -392,7 +393,7 @@ class AWSStack:
         :param sector_name: The name of the sector
         """
         user_config = UserConfig._load_inifile(
-            self.provider.stack[sector_name]["qalx_profile"]
+            self.provider.stack[sector_name]["user_profile"]
         )
         bot_config = BotConfig._load_inifile(
             self.provider.stack[sector_name]["bot_profile"]
